@@ -1,5 +1,5 @@
 /* dw2gencfi.c - Support for generating Dwarf2 CFI information.
-   Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+   Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011
    Free Software Foundation, Inc.
    Contributed by Michal Ludvig <mludvig@suse.cz>
 
@@ -37,7 +37,7 @@
 #endif
 
 #ifndef CFI_DIFF_LSDA_OK
-# define CFI_DIFF_LSDA_OK CFI_DIFF_EXPR_OK
+#define CFI_DIFF_LSDA_OK CFI_DIFF_EXPR_OK
 #endif
 
 #if CFI_DIFF_EXPR_OK == 1 && CFI_DIFF_LSDA_OK == 0
@@ -46,36 +46,206 @@
 
 /* We re-use DWARF2_LINE_MIN_INSN_LENGTH for the code alignment field
    of the CIE.  Default to 1 if not otherwise specified.  */
-#ifndef  DWARF2_LINE_MIN_INSN_LENGTH
-# define DWARF2_LINE_MIN_INSN_LENGTH 1
+#ifndef DWARF2_LINE_MIN_INSN_LENGTH
+#define DWARF2_LINE_MIN_INSN_LENGTH 1
 #endif
 
 /* By default, use 32-bit relocations from .eh_frame into .text.  */
 #ifndef DWARF2_FDE_RELOC_SIZE
-# define DWARF2_FDE_RELOC_SIZE 4
+#define DWARF2_FDE_RELOC_SIZE 4
 #endif
 
 /* By default, use a read-only .eh_frame section.  */
 #ifndef DWARF2_EH_FRAME_READ_ONLY
-# define DWARF2_EH_FRAME_READ_ONLY SEC_READONLY
+#define DWARF2_EH_FRAME_READ_ONLY SEC_READONLY
 #endif
 
 #ifndef EH_FRAME_ALIGNMENT
-# define EH_FRAME_ALIGNMENT (bfd_get_arch_size (stdoutput) == 64 ? 3 : 2)
+#define EH_FRAME_ALIGNMENT (bfd_get_arch_size (stdoutput) == 64 ? 3 : 2)
 #endif
 
 #ifndef tc_cfi_frame_initial_instructions
-# define tc_cfi_frame_initial_instructions() ((void)0)
+#define tc_cfi_frame_initial_instructions() ((void)0)
+#endif
+
+#ifndef tc_cfi_startproc
+# define tc_cfi_startproc() ((void)0)
+#endif
+
+#ifndef tc_cfi_endproc
+# define tc_cfi_endproc(fde) ((void) (fde))
 #endif
 
 #ifndef DWARF2_FORMAT
-# define DWARF2_FORMAT(SEC) dwarf2_format_32bit
+#define DWARF2_FORMAT(SEC) dwarf2_format_32bit
 #endif
 
 #ifndef DWARF2_ADDR_SIZE
-# define DWARF2_ADDR_SIZE(bfd) (bfd_arch_bits_per_address (bfd) / 8)
+#define DWARF2_ADDR_SIZE(bfd) (bfd_arch_bits_per_address (bfd) / 8)
 #endif
 
+#if SUPPORT_FRAME_LINKONCE
+#define CUR_SEG(structp) structp->cur_seg
+#define SET_CUR_SEG(structp, seg) structp->cur_seg = seg 
+#define HANDLED(structp) structp->handled
+#define SET_HANDLED(structp, val) structp->handled = val
+#else
+#define CUR_SEG(structp) NULL
+#define SET_CUR_SEG(structp, seg) (void) (0 && seg)
+#define HANDLED(structp) 0
+#define SET_HANDLED(structp, val) (void) (0 && val)
+#endif
+
+/* Private segment collection list.  */
+struct dwcfi_seg_list
+{
+  segT   seg;
+  int    subseg;
+  char * seg_name;
+};
+
+#define FRAME_NAME ".eh_frame"
+
+static struct hash_control *dwcfi_hash;
+
+/* Build based on segment the derived .debug_...
+   segment name containing origin segment's postfix name part.  */
+
+static char *
+get_debugseg_name (segT seg, const char *base_name)
+{
+  const char *name;
+
+  if (!seg)
+    name = "";
+  else
+    {
+      const char * dollar;
+      const char * dot;
+
+      name = bfd_get_section_name (stdoutput, seg);
+
+      dollar = strchr (name, '$');
+      dot = strchr (name + 1, '.');
+
+      if (!dollar && !dot)
+	name = "";
+      else if (!dollar)
+	name = dot;
+      else if (!dot)
+	name = dollar;
+      else if (dot < dollar)
+	name = dot;
+      else
+	name = dollar;
+    }
+
+  return concat (base_name, name, NULL);
+}
+
+/* Allocate a dwcfi_seg_list structure.  */
+
+static struct dwcfi_seg_list *
+alloc_debugseg_item (segT seg, int subseg, char *name)
+{
+  struct dwcfi_seg_list *r;
+
+  r = (struct dwcfi_seg_list *)
+    xmalloc (sizeof (struct dwcfi_seg_list) + strlen (name));
+  r->seg = seg;
+  r->subseg = subseg;
+  r->seg_name = name;
+  return r;
+}
+
+static segT
+is_now_linkonce_segment (void)
+{
+  if ((bfd_get_section_flags (stdoutput, now_seg)
+       & (SEC_LINK_ONCE | SEC_LINK_DUPLICATES_DISCARD
+	  | SEC_LINK_DUPLICATES_ONE_ONLY | SEC_LINK_DUPLICATES_SAME_SIZE
+	  | SEC_LINK_DUPLICATES_SAME_CONTENTS)) != 0)
+    return now_seg;
+  return NULL;
+}
+
+/* Generate debug... segment with same linkonce properties
+   of based segment.  */
+
+static segT
+make_debug_seg (segT cseg, char *name, int sflags)
+{
+  segT save_seg = now_seg;
+  int save_subseg = now_subseg;
+  segT r;
+  flagword flags;
+
+  r = subseg_new (name, 0);
+
+  /* Check if code segment is marked as linked once.  */
+  if (!cseg)
+    flags = 0;
+  else
+    flags = bfd_get_section_flags (stdoutput, cseg)
+      & (SEC_LINK_ONCE | SEC_LINK_DUPLICATES_DISCARD
+         | SEC_LINK_DUPLICATES_ONE_ONLY | SEC_LINK_DUPLICATES_SAME_SIZE
+         | SEC_LINK_DUPLICATES_SAME_CONTENTS);
+
+  /* Add standard section flags.  */
+  flags |= sflags;
+
+  /* Apply possibly linked once flags to new generated segment, too.  */
+  if (!bfd_set_section_flags (stdoutput, r, flags))
+    as_bad (_("bfd_set_section_flags: %s"),
+	    bfd_errmsg (bfd_get_error ()));
+
+  /* Restore to previous segment.  */
+  if (save_seg != NULL)
+    subseg_set (save_seg, save_subseg);
+  return r;
+}
+
+static void
+dwcfi_hash_insert (const char *name, struct dwcfi_seg_list *item)
+{
+  const char *error_string;
+
+  if ((error_string = hash_jam (dwcfi_hash, name, (char *) item)))
+    as_fatal (_("Inserting \"%s\" into structure table failed: %s"),
+	      name, error_string);
+}
+
+static struct dwcfi_seg_list *
+dwcfi_hash_find (char *name)
+{
+  return (struct dwcfi_seg_list *) hash_find (dwcfi_hash, name);
+}
+
+static struct dwcfi_seg_list *
+dwcfi_hash_find_or_make (segT cseg, const char *base_name, int flags)
+{
+  struct dwcfi_seg_list *item;
+  char *name;
+
+  /* Initialize dwcfi_hash once.  */
+  if (!dwcfi_hash)
+    dwcfi_hash = hash_new ();
+
+  name = get_debugseg_name (cseg, base_name);
+
+  item = dwcfi_hash_find (name);
+  if (!item)
+    {
+      item = alloc_debugseg_item (make_debug_seg (cseg, name, flags), 0, name);
+
+      dwcfi_hash_insert (item->seg_name, item);
+    }
+  else
+    free (name);
+
+  return item;
+}
+
 /* ??? Share this with dwarf2cfg.c.  */
 #ifndef TC_DWARF2_EMIT_OFFSET
 #define TC_DWARF2_EMIT_OFFSET  generic_dwarf2_emit_offset
@@ -94,61 +264,18 @@ generic_dwarf2_emit_offset (symbolS *symbol, unsigned int size)
 }
 #endif
 
-struct cfi_escape_data {
+struct cfi_escape_data
+{
   struct cfi_escape_data *next;
   expressionS exp;
 };
 
-struct cfi_insn_data
-{
-  struct cfi_insn_data *next;
-  int insn;
-  union {
-    struct {
-      unsigned reg;
-      offsetT offset;
-    } ri;
-
-    struct {
-      unsigned reg1;
-      unsigned reg2;
-    } rr;
-
-    unsigned r;
-    offsetT i;
-
-    struct {
-      symbolS *lab1;
-      symbolS *lab2;
-    } ll;
-
-    struct cfi_escape_data *esc;
-
-    struct {
-      unsigned reg, encoding;
-      expressionS exp;
-    } ea;
-  } u;
-};
-
-struct fde_entry
-{
-  struct fde_entry *next;
-  symbolS *start_address;
-  symbolS *end_address;
-  struct cfi_insn_data *data;
-  struct cfi_insn_data **last;
-  unsigned char per_encoding;
-  unsigned char lsda_encoding;
-  expressionS personality;
-  expressionS lsda;
-  unsigned int return_column;
-  unsigned int signal_frame;
-};
-
 struct cie_entry
 {
   struct cie_entry *next;
+#if SUPPORT_FRAME_LINKONCE
+  segT cur_seg;
+#endif
   symbolS *start_address;
   unsigned int return_column;
   unsigned int signal_frame;
@@ -158,8 +285,8 @@ struct cie_entry
   struct cfi_insn_data *first, *last;
 };
 
-
 /* List of FDE entries.  */
+
 static struct fde_entry *all_fde_data;
 static struct fde_entry **last_fde_data = &all_fde_data;
 
@@ -195,7 +322,8 @@ alloc_fde_entry (void)
   frchain_now->frch_cfi_data->cur_fde_data = fde;
   *last_fde_data = fde;
   last_fde_data = &fde->next;
-
+  SET_CUR_SEG (fde, is_now_linkonce_segment ());
+  SET_HANDLED (fde, 0);
   fde->last = &fde->data;
   fde->return_column = DWARF2_DEFAULT_RETURN_COLUMN;
   fde->per_encoding = DW_EH_PE_omit;
@@ -219,13 +347,13 @@ alloc_cfi_insn_data (void)
 
   *cur_fde_data->last = insn;
   cur_fde_data->last = &insn->next;
-
+  SET_CUR_SEG (insn, is_now_linkonce_segment ());
   return insn;
 }
 
 /* Construct a new FDE structure that begins at LABEL.  */
 
-void 
+void
 cfi_new_fde (symbolS *label)
 {
   struct fde_entry *fde = alloc_fde_entry ();
@@ -235,7 +363,7 @@ cfi_new_fde (symbolS *label)
 
 /* End the currently open FDE.  */
 
-void 
+void
 cfi_end_fde (symbolS *label)
 {
   frchain_now->frch_cfi_data->cur_fde_data->end_address = label;
@@ -254,7 +382,7 @@ cfi_set_return_column (unsigned regno)
 /* Universal functions to store new instructions.  */
 
 static void
-cfi_add_CFA_insn(int insn)
+cfi_add_CFA_insn (int insn)
 {
   struct cfi_insn_data *insn_ptr = alloc_cfi_insn_data ();
 
@@ -424,14 +552,6 @@ static void dot_cfi_personality (int);
 static void dot_cfi_lsda (int);
 static void dot_cfi_val_encoded_addr (int);
 
-/* Fake CFI type; outside the byte range of any real CFI insn.  */
-#define CFI_adjust_cfa_offset	0x100
-#define CFI_return_column	0x101
-#define CFI_rel_offset		0x102
-#define CFI_escape		0x103
-#define CFI_signal_frame	0x104
-#define CFI_val_encoded_addr	0x105
-
 const pseudo_typeS cfi_pseudo_table[] =
   {
     { "cfi_sections", dot_cfi_sections, 0 },
@@ -471,7 +591,7 @@ cfi_parse_separator (void)
 
 #ifndef tc_parse_to_dw2regnum
 static void
-tc_parse_to_dw2regnum(expressionS *exp)
+tc_parse_to_dw2regnum (expressionS *exp)
 {
 # ifdef tc_regname_to_dw2regnum
   SKIP_WHITESPACE ();
@@ -853,7 +973,7 @@ dot_cfi_val_encoded_addr (int ignored ATTRIBUTE_UNUSED)
 
   insn_ptr = alloc_cfi_insn_data ();
   insn_ptr->insn = CFI_val_encoded_addr;
-  
+
   insn_ptr->u.ea.reg = cfi_parse_reg ();
 
   cfi_parse_separator ();
@@ -900,6 +1020,7 @@ dot_cfi_val_encoded_addr (int ignored ATTRIBUTE_UNUSED)
 /* By default emit .eh_frame only, not .debug_frame.  */
 #define CFI_EMIT_eh_frame	(1 << 0)
 #define CFI_EMIT_debug_frame	(1 << 1)
+#define CFI_EMIT_target		(1 << 2)
 static int cfi_sections = CFI_EMIT_eh_frame;
 
 static void
@@ -916,10 +1037,15 @@ dot_cfi_sections (int ignored ATTRIBUTE_UNUSED)
 	name = input_line_pointer;
 	c = get_symbol_end ();
 
-	if (strcmp (name, ".eh_frame") == 0)
+	if (strncmp (name, ".eh_frame", sizeof ".eh_frame") == 0
+	    && name[9] != '_')
 	  sections |= CFI_EMIT_eh_frame;
-	else if (strcmp (name, ".debug_frame") == 0)
+	else if (strncmp (name, ".debug_frame", sizeof ".debug_frame") == 0)
 	  sections |= CFI_EMIT_debug_frame;
+#ifdef tc_cfi_section_name
+	else if (strcmp (name, tc_cfi_section_name) == 0)
+	  sections |= CFI_EMIT_target;
+#endif
 	else
 	  {
 	    *input_line_pointer = c;
@@ -982,11 +1108,16 @@ dot_cfi_startproc (int ignored ATTRIBUTE_UNUSED)
   frchain_now->frch_cfi_data->cur_cfa_offset = 0;
   if (!simple)
     tc_cfi_frame_initial_instructions ();
+
+  if ((cfi_sections & CFI_EMIT_target) != 0)
+    tc_cfi_startproc ();
 }
 
 static void
 dot_cfi_endproc (int ignored ATTRIBUTE_UNUSED)
 {
+  struct fde_entry *fde;
+
   if (frchain_now->frch_cfi_data == NULL)
     {
       as_bad (_(".cfi_endproc without corresponding .cfi_startproc"));
@@ -994,9 +1125,14 @@ dot_cfi_endproc (int ignored ATTRIBUTE_UNUSED)
       return;
     }
 
+  fde = frchain_now->frch_cfi_data->cur_fde_data;
+
   cfi_end_fde (symbol_temp_new_now ());
 
   demand_empty_rest_of_line ();
+
+  if ((cfi_sections & CFI_EMIT_target) != 0)
+    tc_cfi_endproc (fde);
 }
 
 
@@ -1250,7 +1386,7 @@ output_cfi_insn (struct cfi_insn_data *insn)
 	emit_expr (&insn->u.ea.exp, encoding_size);
       }
       break;
-      
+
     default:
       abort ();
     }
@@ -1384,8 +1520,14 @@ output_cie (struct cie_entry *cie, bfd_boolean eh_frame, int align)
     out_one (enc);
 
   if (cie->first)
-    for (i = cie->first; i != cie->last; i = i->next)
-      output_cfi_insn (i);
+    {
+      for (i = cie->first; i != cie->last; i = i->next)
+        {
+	  if (CUR_SEG (i) != CUR_SEG (cie))
+	    continue;
+	  output_cfi_insn (i);
+	}
+    }
 
   frag_align (align, DW_CFA_nop, 0);
   symbol_set_value_now (end_address);
@@ -1492,7 +1634,8 @@ output_fde (struct fde_entry *fde, struct cie_entry *cie,
     }
 
   for (; first; first = first->next)
-    output_cfi_insn (first);
+    if (CUR_SEG (first) == CUR_SEG (fde))
+      output_cfi_insn (first);
 
   frag_align (align, DW_CFA_nop, 0);
   symbol_set_value_now (end_address);
@@ -1507,6 +1650,8 @@ select_cie_for_fde (struct fde_entry *fde, bfd_boolean eh_frame,
 
   for (cie = cie_root; cie; cie = cie->next)
     {
+      if (CUR_SEG (cie) != CUR_SEG (fde))
+	continue;
       if (cie->return_column != fde->return_column
 	  || cie->signal_frame != fde->signal_frame
 	  || cie->per_encoding != fde->per_encoding
@@ -1605,6 +1750,7 @@ select_cie_for_fde (struct fde_entry *fde, bfd_boolean eh_frame,
   cie = (struct cie_entry *) xmalloc (sizeof (struct cie_entry));
   cie->next = cie_root;
   cie_root = cie;
+  SET_CUR_SEG (cie, CUR_SEG (fde));
   cie->return_column = fde->return_column;
   cie->signal_frame = fde->signal_frame;
   cie->per_encoding = fde->per_encoding;
@@ -1629,130 +1775,227 @@ select_cie_for_fde (struct fde_entry *fde, bfd_boolean eh_frame,
 
 #ifdef md_reg_eh_frame_to_debug_frame
 static void
-cfi_change_reg_numbers (struct cfi_insn_data *insn)
+cfi_change_reg_numbers (struct cfi_insn_data *insn, segT ccseg)
 {
   for (; insn; insn = insn->next)
-    switch (insn->insn)
-      {
-      case DW_CFA_advance_loc:
-      case DW_CFA_def_cfa_offset:
-      case DW_CFA_remember_state:
-      case DW_CFA_restore_state:
-      case DW_CFA_GNU_window_save:
-      case CFI_escape:
-	break;
+    {
+      if (CUR_SEG (insn) != ccseg)
+        continue;
+      switch (insn->insn)
+	{
+	case DW_CFA_advance_loc:
+	case DW_CFA_def_cfa_offset:
+	case DW_CFA_remember_state:
+	case DW_CFA_restore_state:
+	case DW_CFA_GNU_window_save:
+	case CFI_escape:
+	  break;
 
-      case DW_CFA_def_cfa:
-      case DW_CFA_offset:
-	insn->u.ri.reg = md_reg_eh_frame_to_debug_frame (insn->u.ri.reg);
-	break;
+	case DW_CFA_def_cfa:
+	case DW_CFA_offset:
+	  insn->u.ri.reg = md_reg_eh_frame_to_debug_frame (insn->u.ri.reg);
+	  break;
 
-      case DW_CFA_def_cfa_register:
-      case DW_CFA_undefined:
-      case DW_CFA_same_value:
-      case DW_CFA_restore:
-	insn->u.r = md_reg_eh_frame_to_debug_frame (insn->u.r);
-	break;
+	case DW_CFA_def_cfa_register:
+	case DW_CFA_undefined:
+	case DW_CFA_same_value:
+	case DW_CFA_restore:
+	  insn->u.r = md_reg_eh_frame_to_debug_frame (insn->u.r);
+	  break;
 
-      case DW_CFA_register:
-	insn->u.rr.reg1 = md_reg_eh_frame_to_debug_frame (insn->u.rr.reg1);
-	insn->u.rr.reg2 = md_reg_eh_frame_to_debug_frame (insn->u.rr.reg2);
-	break;
+	case DW_CFA_register:
+	  insn->u.rr.reg1 = md_reg_eh_frame_to_debug_frame (insn->u.rr.reg1);
+	  insn->u.rr.reg2 = md_reg_eh_frame_to_debug_frame (insn->u.rr.reg2);
+	  break;
 
-      case CFI_val_encoded_addr:
-	insn->u.ea.reg = md_reg_eh_frame_to_debug_frame (insn->u.ea.reg);
-	break;
+	case CFI_val_encoded_addr:
+	  insn->u.ea.reg = md_reg_eh_frame_to_debug_frame (insn->u.ea.reg);
+	  break;
 
-      default:
-	abort ();
-      }
+	default:
+	  abort ();
+	}
+    }
 }
 #else
-#define cfi_change_reg_numbers(insn) do { } while (0)
+#define cfi_change_reg_numbers(insn, cseg) do { } while (0)
 #endif
 
+static segT
+get_cfi_seg (segT cseg, const char *base, flagword flags, int align)
+{
+  if (SUPPORT_FRAME_LINKONCE)
+    {
+      struct dwcfi_seg_list *l;
+
+      l = dwcfi_hash_find_or_make (cseg, base, flags);
+
+      cseg = l->seg;
+      subseg_set (cseg, l->subseg);
+    }
+  else
+    {
+      cseg = subseg_new (base, 0);
+      bfd_set_section_flags (stdoutput, cseg, flags);
+    }
+  record_alignment (cseg, align);
+  return cseg;
+}
+
 void
 cfi_finish (void)
 {
-  segT cfi_seg;
+  struct cie_entry *cie, *cie_next;
+  segT cfi_seg, ccseg;
   struct fde_entry *fde;
-  int save_flag_traditional_format;
+  struct cfi_insn_data *first;
+  int save_flag_traditional_format, seek_next_seg;
 
   if (all_fde_data == 0)
     return;
 
   if ((cfi_sections & CFI_EMIT_eh_frame) != 0)
     {
-      /* Open .eh_frame section.  */
-      cfi_seg = subseg_new (".eh_frame", 0);
-      bfd_set_section_flags (stdoutput, cfi_seg,
-			     SEC_ALLOC | SEC_LOAD | SEC_DATA
-			     | DWARF2_EH_FRAME_READ_ONLY);
-      subseg_set (cfi_seg, 0);
-      record_alignment (cfi_seg, EH_FRAME_ALIGNMENT);
-
-#ifdef md_fix_up_eh_frame
-      md_fix_up_eh_frame (cfi_seg);
-#endif
-
       /* Make sure check_eh_frame doesn't do anything with our output.  */
       save_flag_traditional_format = flag_traditional_format;
       flag_traditional_format = 1;
 
-      for (fde = all_fde_data; fde ; fde = fde->next)
+      if (!SUPPORT_FRAME_LINKONCE)
 	{
-	  struct cfi_insn_data *first;
-	  struct cie_entry *cie;
-
-	  if (fde->end_address == NULL)
-	    {
-	      as_bad (_("open CFI at the end of file; missing .cfi_endproc directive"));
-	      fde->end_address = fde->start_address;
-	    }
-
-	  cie = select_cie_for_fde (fde, TRUE, &first, 2);
-	  output_fde (fde, cie, TRUE, first,
-		      fde->next == NULL ? EH_FRAME_ALIGNMENT : 2);
+	  /* Open .eh_frame section.  */
+	  cfi_seg = get_cfi_seg (NULL, ".eh_frame",
+				 (SEC_ALLOC | SEC_LOAD | SEC_DATA
+				  | DWARF2_EH_FRAME_READ_ONLY),
+				 EH_FRAME_ALIGNMENT);
+#ifdef md_fix_up_eh_frame
+	  md_fix_up_eh_frame (cfi_seg);
+#else
+	  (void) cfi_seg;
+#endif
 	}
 
+      do
+        {
+	  ccseg = NULL;
+	  seek_next_seg = 0;
+
+	  for (cie = cie_root; cie; cie = cie_next)
+	    {
+	      cie_next = cie->next;
+	      free ((void *) cie);
+	    }
+	  cie_root = NULL;
+
+	  for (fde = all_fde_data; fde ; fde = fde->next)
+	    {
+	      if (SUPPORT_FRAME_LINKONCE)
+		{
+		  if (HANDLED (fde))
+		    continue;
+		  if (seek_next_seg && CUR_SEG (fde) != ccseg)
+		    {
+		      seek_next_seg = 2;
+		      continue;
+		    }
+		  if (!seek_next_seg)
+		    {
+		      ccseg = CUR_SEG (fde);
+		      /* Open .eh_frame section.  */
+		      cfi_seg = get_cfi_seg (ccseg, ".eh_frame",
+					     (SEC_ALLOC | SEC_LOAD | SEC_DATA
+					      | DWARF2_EH_FRAME_READ_ONLY),
+					     EH_FRAME_ALIGNMENT);
+#ifdef md_fix_up_eh_frame
+		      md_fix_up_eh_frame (cfi_seg);
+#else
+		      (void) cfi_seg;
+#endif
+		      seek_next_seg = 1;
+		    }
+		  SET_HANDLED (fde, 1);
+		}
+
+	      if (fde->end_address == NULL)
+		{
+		  as_bad (_("open CFI at the end of file; missing .cfi_endproc directive"));
+		  fde->end_address = fde->start_address;
+		}
+
+	      cie = select_cie_for_fde (fde, TRUE, &first, 2);
+	      output_fde (fde, cie, TRUE, first,
+			  fde->next == NULL ? EH_FRAME_ALIGNMENT : 2);
+	    }
+	}
+      while (SUPPORT_FRAME_LINKONCE && seek_next_seg == 2);
+
+      if (SUPPORT_FRAME_LINKONCE)
+	for (fde = all_fde_data; fde ; fde = fde->next)
+	  SET_HANDLED (fde, 0);
+
       flag_traditional_format = save_flag_traditional_format;
     }
 
   if ((cfi_sections & CFI_EMIT_debug_frame) != 0)
     {
-      struct cie_entry *cie, *cie_next;
       int alignment = ffs (DWARF2_ADDR_SIZE (stdoutput)) - 1;
 
-      for (cie = cie_root; cie; cie = cie_next)
-	{
-	  cie_next = cie->next;
-	  free ((void *) cie);
-	}
-      cie_root = NULL;
+      if (!SUPPORT_FRAME_LINKONCE)
+	get_cfi_seg (NULL, ".debug_frame",
+		     SEC_READONLY | SEC_DEBUGGING,
+		     alignment);
 
-      /* Open .debug_frame section.  */
-      cfi_seg = subseg_new (".debug_frame", 0);
-      bfd_set_section_flags (stdoutput, cfi_seg,
-			     SEC_READONLY | SEC_DEBUGGING);
-      subseg_set (cfi_seg, 0);
-      record_alignment (cfi_seg, alignment);
+      do
+        {
+	  ccseg = NULL;
+	  seek_next_seg = 0;
 
-      for (fde = all_fde_data; fde ; fde = fde->next)
-	{
-	  struct cfi_insn_data *first;
-
-	  if (fde->end_address == NULL)
+	  for (cie = cie_root; cie; cie = cie_next)
 	    {
-	      as_bad (_("open CFI at the end of file; missing .cfi_endproc directive"));
-	      fde->end_address = fde->start_address;
+	      cie_next = cie->next;
+	      free ((void *) cie);
 	    }
+	  cie_root = NULL;
 
-	  fde->per_encoding = DW_EH_PE_omit;
-	  fde->lsda_encoding = DW_EH_PE_omit;
-	  cfi_change_reg_numbers (fde->data);
-	  cie = select_cie_for_fde (fde, FALSE, &first, alignment);
-	  output_fde (fde, cie, FALSE, first, alignment);
+	  for (fde = all_fde_data; fde ; fde = fde->next)
+	    {
+	      if (SUPPORT_FRAME_LINKONCE)
+		{
+		  if (HANDLED (fde))
+		    continue;
+		  if (seek_next_seg && CUR_SEG (fde) != ccseg)
+		    {
+		      seek_next_seg = 2;
+		      continue;
+		    }
+		  if (!seek_next_seg)
+		    {
+		      ccseg = CUR_SEG (fde);
+		      /* Open .debug_frame section.  */
+		      get_cfi_seg (ccseg, ".debug_frame",
+				   SEC_READONLY | SEC_DEBUGGING,
+				   alignment);
+		      seek_next_seg = 1;
+		    }
+		  SET_HANDLED (fde, 1);
+		}
+	      if (fde->end_address == NULL)
+		{
+		  as_bad (_("open CFI at the end of file; missing .cfi_endproc directive"));
+		  fde->end_address = fde->start_address;
+		}
+
+	      fde->per_encoding = DW_EH_PE_omit;
+	      fde->lsda_encoding = DW_EH_PE_omit;
+	      cfi_change_reg_numbers (fde->data, ccseg);
+	      cie = select_cie_for_fde (fde, FALSE, &first, alignment);
+	      output_fde (fde, cie, FALSE, first, alignment);
+	    }
 	}
+      while (SUPPORT_FRAME_LINKONCE && seek_next_seg == 2);
+
+      if (SUPPORT_FRAME_LINKONCE)
+	for (fde = all_fde_data; fde ; fde = fde->next)
+	  SET_HANDLED (fde, 0);
     }
 }
 
